//-----------------------------------------------------------------------------
//  Copyright (C) 2014-2022 Thomas S. Ullrich
//
//  This file is part of "MineSweeper".
//
//  This file may be used under the terms of the GNU General Public License.
//  This project is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License.
//
//  Author: Thomas S. Ullrich
//  Last update: April 18, 2022
//-----------------------------------------------------------------------------
#include "msAboutDialog.h"
#include "msVersion.h"
#include <iostream>
#include <QtGui>
#include <QFrame>
#include <QPushButton>
#include <QPushButton>
#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QScreen>
#include <QGuiApplication>

using namespace std; 

#define PR(x) cout << #x << " = " << endl;

msAboutDialog::msAboutDialog(QWidget* parent)
    : QDialog(parent),
      mTitlePixmap(":/images/msSplash.png"),
      mVersionText(QString("Version ") + QString(VERSION) + QString(" (build ") + QString(__DATE__) + QString(")"))
{
    setModal(true);

    //
    //  Get screen
    //
    QScreen *screen = QGuiApplication::primaryScreen();

#if defined(Q_OS_MAC)
    //
    //  Change icon if Retina display (Mac)
    //
    if (screen->devicePixelRatio() > 1) {
        mTitlePixmap.load(":/images/msSplash@2x.png");
        mTitlePixmap.setDevicePixelRatio(2.0);
    }
#endif
    
    //
    //  Merge version/release string into pixmap
    //
#if defined(Q_OS_MAC)
    QFont font("Helvetica", 15);
#else
    QFont font("Arial", 11);
#endif
    QPainter p;
    p.begin( &mTitlePixmap ); 
    p.setPen(Qt::black);
    p.setFont(font);
    int strWidth = p.fontMetrics().horizontalAdvance( mVersionText );
    int strHeight = p.fontMetrics().height(); 
    p.drawText(QRect(300, 70, strWidth, strHeight), 0, mVersionText );
    mCopyrightText = QString("Copyright (C) 2014-2023 Thomas Ullrich");
    strWidth = p.fontMetrics().horizontalAdvance( mCopyrightText );
    p.drawText(QRect(300, 65+1.6*strHeight, strWidth, strHeight), 0, mCopyrightText );
    p.end();
    
    //
    //  First define the pixmap label
    //
    int w = mTitlePixmap.width()/screen->devicePixelRatio() ;
    int h = mTitlePixmap.height()/screen->devicePixelRatio() ;
    
    mTitlePixmapLabel = new QLabel(this);

    mTitlePixmapLabel->setGeometry(QRect( 0, 0, w, h));
    mTitlePixmapLabel->setFixedSize(QSize(w, h));
    mTitlePixmapLabel->setLineWidth(0);
	mTitlePixmapLabel->setPixmap(mTitlePixmap);
    mTitleShown = true;

    mTitlePixmapLabel->setScaledContents(false);
    mTitlePixmapLabel->setAlignment(Qt::AlignLeft|Qt::AlignTop);
    mTitlePixmapLabel->setIndent(10); // for text

    //
    //  Now the line (just looks nicer with it)
    //
    mLine = new QFrame(this);
    mLine->setGeometry(QRect(0, h, w, 5));
    mLine->setFrameShape(QFrame::HLine);
    mLine->setFrameShadow(QFrame::Sunken);

    //
    //  The buttons
    //
    mButtonOk = new QPushButton(this);
    mButtonOk->setAutoDefault(true);
    mButtonOk->setDefault(true);
    mButtonOk->setText( tr("&Ok" ));
    mButtonLicense = new QPushButton(this);
    mButtonLicense->setText(tr("&License"));

    //
    //  Then the dialog itself
    //
    setWindowTitle(tr( "About MineSweeper"));
    setSizeGripEnabled(false);
    setFixedWidth(w);

    //
    //  Layout the buttons
    //
    QSpacerItem* spacer = new QSpacerItem (300,20, QSizePolicy::Expanding, QSizePolicy::Minimum );
    mButtonLayout = new QHBoxLayout;
    mButtonLayout->addWidget(mButtonLicense);
    mButtonLayout->addItem(spacer);
    mButtonLayout->addWidget(mButtonOk);
    mButtonLayout->setContentsMargins(10, 0, 10, 0);

    //
    //  Overall widget layout
    //
	mLayout = new QVBoxLayout(this);
	mLayout->setSpacing(0);
    mLayout->setContentsMargins(0, 0, 0, 0);
    mLayout->addWidget(mTitlePixmapLabel);
    mLayout->addWidget(mLine);
    mLayout->addLayout(mButtonLayout);

    setFixedSize(mLayout->minimumSize());
   
    //
    //  Connect slots
    //
    connect( mButtonOk, SIGNAL(clicked()), this, SLOT(accept()));
    connect( mButtonLicense, SIGNAL(clicked()), this, SLOT(license()));
}

void msAboutDialog::accept()  // Ok button pressed
{
    //
    // Hide it and make sure the next time it pops
    // up the title page is shown.
    //
    hide();
    if (!mTitleShown) license();
}

void msAboutDialog::license()  // License button pressed
{
    if (mTitleShown) {
        QPalette palette;
        palette.setColor(mTitlePixmapLabel->backgroundRole(), Qt::white);// white background
        mTitlePixmapLabel->setPalette(palette);       
        mButtonLicense->setText(tr("Title && &Version"));
        mTitleShown = false;
        mTitlePixmapLabel->setText(tr(
            "Copyright (C) 2014-2023 Thomas S. Ullrich\n"
            "\n"
            "This program is free software; you can redistribute it and/or modify it under the terms\n"
            "of the GNU General Public License as published by the Free Software Foundation; either\n"
            "version 3 of the License, or any later version.\n"
            "\n"
            "This program is distributed  in the hope that it will be useful, but without any warranty;\n"
            "without even the implied warranty of merchantability or fitness for a particular purpose. \n"
            "See the GNU General Public License for more details.\n"
            "\n"
            "You should have received a copy of the GNU General Public License along with this program.\n"
            "If not, see <http://www.gnu.org/licenses/>."
            ));
    }
    else {
        mTitlePixmapLabel->setPixmap(mTitlePixmap);	
        mButtonLicense->setText(tr("&License"));
        mTitleShown = true;
    }
    update();
}


