//-----------------------------------------------------------------------------
//  Copyright (C) 2014-2021 Thomas S. Ullrich
//
//  This file is part of "MineSweeper".
//
//  This file may be used under the terms of the GNU General Public License.
//  This project is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License.
//
//  Author: Thomas S. Ullrich
//  Last update: March 25, 2021
//-----------------------------------------------------------------------------
#include <QtGui>
#include <QGraphicsView>
#include <QGraphicsScene>
#include <QGraphicsRectItem>
#include <QGraphicsSvgItem>
#include <QGraphicsSceneMouseEvent>
#include <QLabel>
#include <QStatusBar>
#include <QMenuBar>
#include <QApplication>
#include <QLayout>
#include <QSvgRenderer>
#include <QInputDialog>
#include <QMessageBox>
#include <cmath>
#include <iostream>
#include "msWindow.h"
#include "msCell.h"
#include "msCustomSizeDialog.h"
#include "msPreferencesDialog.h"
#include "msSettings.h"
#include "msHighscoreList.h"
#include "msAboutDialog.h"
#include "msHelpBrowser.h"
#include "msOopsPixmap.h"

using namespace std;

#define PR(x) cout << #x << " = " << (x) << endl;

msWindow::msWindow()
{
    setWindowTitle("MineSweeper");
    setAttribute(Qt::WA_QuitOnClose);
    mOopsiesLeft = 0;
    
    //
    //  Init time counter
    //
    mStartTime = 0;
    mEndTime = 0;
    mInactiveTime = 0;

    //
    //  Setup random engine
    //
    random_device rd;
    default_random_engine newEngine(rd());
    mRandomEngine = newEngine;

    //
    //  Load settings first, they are needed early
    //
    mHighscoreList = 0;
    mSettings = new msSettings();
    mSettings->load();  // loads preferences but not high scores

    //
    //  Define Menus & Actions
    //

    //
    //  Game Menu
    //
    mNewGameAction = new QAction(tr("&New Game"), this);
    mNewGameAction->setShortcut(QKeySequence::New);
    mHighscoreAction = new QAction(tr("&High Scores..."), this);
    // No shutcut for high scores
    mPreferencesAction = new QAction("Preferences", this);
    mPreferencesAction->setShortcut(QKeySequence::Preferences);
    mFinishAction = new QAction("Close", this);
    mFinishAction->setShortcut(QKeySequence::Close);
 
    mGameMenu = menuBar()->addMenu(tr("&Game"));
    mGameMenu->addAction(mNewGameAction);
    mGameMenu->addAction(mHighscoreAction);
    mGameMenu->addAction(mPreferencesAction);
    mGameMenu->addSeparator();
    mGameMenu->addAction(mFinishAction);

    //
    //  Game Size Menu
    //
    mSmallSizeAction = new QAction(tr("Small"), this);
    mSmallSizeAction->setShortcut(Qt::CTRL | Qt::Key_1);
    mMediumSizeAction = new QAction(tr("Medium"), this);
    mMediumSizeAction->setShortcut(Qt::CTRL | Qt::Key_2);
    mLargeSizeAction = new QAction(tr("Large"), this);
    mLargeSizeAction->setShortcut(Qt::CTRL | Qt::Key_3);
    mCustomSizeAction = new QAction(tr("Custom ..."), this);
    mCustomSizeAction->setShortcut(QKeySequence::Copy);
    
    mGameSizeMenu = menuBar()->addMenu(tr("Game &Size"));
    mGameSizeMenu->addAction(mSmallSizeAction);
    mGameSizeMenu->addAction(mMediumSizeAction);
    mGameSizeMenu->addAction(mLargeSizeAction);
    mGameSizeMenu->addSeparator();
    mGameSizeMenu->addAction(mCustomSizeAction);

    //
    //  Help Menu
    //
    mAboutAction = new QAction(tr("&About MineSweeper"), this);
    mHelpAction = new QAction(tr("MineSweeper &Help"), this);
    connect(mAboutAction, SIGNAL(triggered()), this, SLOT(about()));
    connect(mHelpAction, SIGNAL(triggered()), this, SLOT(help()));
#if defined(Q_OS_MAC)
    mAboutAction->setMenuRole(QAction::AboutRole);
#endif
    mHelpMenu = menuBar()->addMenu(tr("&Help"));
    mHelpMenu->addAction(mHelpAction);
#if !defined(Q_OS_MAC)
    mHelpMenu->addSeparator();
#endif
    mHelpMenu->addAction(mAboutAction);  // In application menu on MacOS/X
    
    //
    //  Connect signals and slots, part I
    //
    connect(mNewGameAction, SIGNAL(triggered()), this, SLOT(startGame()));
    connect(mSmallSizeAction, SIGNAL(triggered()), this, SLOT(setSmallSizeAndStartNew()));
    connect(mMediumSizeAction, SIGNAL(triggered()), this, SLOT(setMediumSizeAndStartNew()));
    connect(mLargeSizeAction, SIGNAL(triggered()), this, SLOT(setLargeSizeAndStartNew()));
    connect(mCustomSizeAction, SIGNAL(triggered()), this, SLOT(setCustomSize()));
    connect(mPreferencesAction, SIGNAL(triggered()), this, SLOT(showPreferences()));
    connect(mHighscoreAction, SIGNAL(triggered()), this, SLOT(showHighscores()));
    connect(mFinishAction, SIGNAL(triggered()), this, SLOT(finish()));

    //
    //   Setup graphics view and scene
    //
    mGraphicsView = new QGraphicsView;
    mGraphicsScene = new QGraphicsScene(mGraphicsView);
    mGraphicsView->setScene(mGraphicsScene);
    mGraphicsView->setCacheMode(QGraphicsView::CacheBackground);
    mGraphicsView->setFrameStyle(QFrame::NoFrame);
    mGraphicsView->setViewportUpdateMode(QGraphicsView::FullViewportUpdate);
    mGraphicsView->setAcceptDrops(false);
    mGraphicsView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    mGraphicsView->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    mGraphicsView->setRenderHint(QPainter::Antialiasing, true);
    setCentralWidget(mGraphicsView);
    mGraphicsScene->installEventFilter(this);
    mGraphicsScene->setBackgroundBrush(msSettings::mBackgroundColor);

    //
    //  Setup status bar and timer
    //
    createStatusBar();
    mClockUpdateTimer = new QTimer(this);

    //
    //   Create dialogs
    //
    mAboutDialog = new msAboutDialog(this);
    mCustomSizeDialog = new msCustomSizeDialog(this);
    mCustomSizeDialog->setMaximumRows(msSettings::mMaxNumYCells);
    mCustomSizeDialog->setMaximumColumns(msSettings::mMaxNumXCells);
    mPreferencesDialog = new msPreferencesDialog(this);
    mHighscoreList = new msHighscoreList(this);
    mSettings->setHighscoreList(mHighscoreList);
    createHelpBrowser();

    //
    //  Define and set limits for mCustomSizeDialog
    //  and mPreferenceDialog.
    //
    ensureBoardFitsOnScreen();
    
    //
    //  Create SVG Renderers
    //
    mFlagSvgRenderer = new QSvgRenderer(msSettings::flagFileName());
    mMineSvgRenderer = new QSvgRenderer(msSettings::mineFileName());
    mTryOutSvgRenderer = new QSvgRenderer(msSettings::tryOutSymbolFileName());
    const int maxNeighbors = 8;
    for (int i=1; i<=maxNeighbors; i++)
        mNumberSvgRenderer[i] = new QSvgRenderer(msSettings::numberFileName(i));

    //
    //  Setup Oops animation
    //
    QPixmap oops(":/images/oops.png");
    double dpi = QGuiApplication::primaryScreen()->devicePixelRatio();
    oops.setDevicePixelRatio(dpi);    // retina
    mOopsPixmap = new msOopsPixmap;
    mOopsPixmap->setPixmap(oops);
    mOopsPixmap->setTransformationMode(Qt::SmoothTransformation); // fast enough but nicer animated
    mAnimation = new QPropertyAnimation(mOopsPixmap, "scale");
    mAnimation->setDuration(800);   // msec
    mAnimation->setKeyValueAt(0, 0.);
    mAnimation->setKeyValueAt(0.3, 0.3*dpi);
    mAnimation->setKeyValueAt(0.9, 0.3*dpi);
    mAnimation->setKeyValueAt(1, 0.0);
    
    //
	//  Connect signals and slots, part II
	//
    connect(mClockUpdateTimer, SIGNAL(timeout()), this, SLOT(updateClock()));
    connect(mCustomSizeDialog, SIGNAL(finished(int)), this, SLOT(customSizeSelected(int )));
    connect(mPreferencesDialog, SIGNAL(finished(int)), this, SLOT(newPreferences(int)));
    connect(mAnimation, SIGNAL(finished()), this, SLOT(finishOopsAnimation()));
  
    //
    //   Load preferences again, this time with highscores
    //
    mSettings->load();

    //
    //   Set icon for all windows
    //
    setWindowIcon(QIcon(QString::fromUtf8(":/images/minesweeperIcon.png")));

    startGame();

#if defined(Q_OS_WIN)
    //
    // In older versions this was not necessary but the
    // mechanism to calculate the correct size of the main
    // window in Windows is broken (as it is on X11).
    // Only once the window is actually visible can the
    // correct size be calculated.
    //
    show();
    resizeMainWindow();
#endif
}

msWindow::~msWindow()
{
    mSettings->save();
    delete mSettings;
    delete mPreferencesDialog;
    delete mHighscoreList;
    delete mAboutDialog;
    delete mCustomSizeDialog;
}

void msWindow::about()
{
    mAboutDialog->show();
}

void msWindow::changeEvent(QEvent* e)
{
    static bool wasMinimized = false;
    if (e->type() == QEvent::WindowStateChange) {
        if (isMinimized() && !wasMinimized) {
            // main window just got iconified
            if (mGameIsOn) mInactiveTime = time(0);
        }
        else {
            // main window just became visible after being iconified
            if (mGameIsOn && mStartTime != 0) {
                time_t delta = mInactiveTime > 0 ? (time(0) - mInactiveTime) : 0;
                mStartTime += delta;
                mInactiveTime = 0;
            }
        }
        wasMinimized = isMinimized();
    }
    if (mGameIsOn) updateClock();
    
    QMainWindow::changeEvent(e);
}

void msWindow::clearObvious(unsigned int ix, unsigned int iy)
{
    if (mCellArray[ix][iy]->numberOfNeighbors() != 0) return;
    
    for (int dy=-1; dy<=1; dy++) {
        int y = iy+dy;
        if (y < 0 || y >= static_cast<int>(msSettings::mNumYCells)) continue;
        for (int dx=-1; dx<=1; dx++) {
            int x = ix+dx;
            if (x < 0 || x >= static_cast<int>(msSettings::mNumXCells)) continue;
            if (!mCellArray[x][y]->isRevealed()) {  // needed to stop recursion
                mCellArray[x][y]->setIsRevealed(true);
                mCellArray[x][y]->waterRectangle()->hide();
                mCellArray[x][y]->flagSvg()->hide();
                mCellArray[x][y]->tryOutText()->hide();
                if (mCellArray[x][y]->numberOfNeighbors() == 0) clearObvious(x, y);
            }
        }
    }
}

void msWindow::createHelpBrowser()
{
    mHelpBrowser = 0;

    //
    //  The documentation is in the docs/ directory.
    //  On MacOS it is the app bundle under Resources,
    //  on Linux typically in /usr/local/share/MineSweeper.
    //  Here we go through the various options step by step.
    //
    
    QString path;
    
    //
    //  1. This is for the developer using MineSweeper/docs directly
    //  for easier testing and turn around times.
    //
    path = "/Users/ullrich/Documents/Projects/MineSweeper/docs";
    QDir dir(path);
    
    //
    //  2. MacOS
    //
    if (!dir.exists()) {
        path = qApp->applicationDirPath() + "/../Resources/docs";
        dir.setPath(path);
    }
  
    //
    //  3. Windows
    //
    if (!dir.exists()) {
        path = qApp->applicationDirPath() + "/docs";
        dir.setPath(path);
    }
    if (!dir.exists()) {
        path = qApp->applicationDirPath() + "/../../docs";
        dir.setPath(path);
    }

    //
    //  4. Linux
    //
    if (!dir.exists()) {
        path = "/usr/local/share/MineSweeper/docs";
        dir.setPath(path);
    }
    
    //
    //  5. Just in case we check other locations if nothing
    //     was found so far. More an act of desperation.
    //
    if (!dir.exists()) {
        path = qApp->applicationDirPath() + "/docs";
        dir.setPath(path);
    }
    if (!dir.exists()) {
        path = qApp->applicationDirPath() + "/../docs";
        dir.setPath(path);
    }
    if (!dir.exists()) {
        path = qApp->applicationDirPath() + "../../docs";
        dir.setPath(path);
    }
    if (!dir.exists()) {
        path = "/usr/local/MineSweeper/docs";
        dir.setPath(path);
    }
    if (!dir.exists()) {
        path = "/opt/local/MineSweeper/docs";
        dir.setPath(path);
    }
    
    if (!dir.exists()) {
        QMessageBox::warning(0, "MineSweeper",
                             tr("Cannot find the directory holding the documentation (docs). "
                                "No online help will be available. "
                                "Check your installation and reinstall if necessary."));
        return;
    }

    //
    //  Setup the help browser
    //

    QString descFile = path + "/doc.index";
    
    QFileInfo file(descFile);
    if (file.exists())
        mHelpBrowser = new msHelpBrowser(descFile, path);
    else {
        QMessageBox::warning( 0, "MineSweeper",
                             tr("Cannot find the index descriptor to setup the documentation. "
                                "No online help will be available. "
                                "Check your installation and reinstall if necessary."));
    }
}

void msWindow::createStatusBar() {
    
    mFlagCountLabel = new QLabel("000/000", this);
    mFlagCountLabel->setAlignment(Qt::AlignHCenter);
    mFlagCountLabel->setMinimumWidth(statusBar()->fontMetrics().horizontalAdvance(" 999/999 "));
    mFlagCountLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    
    mMessageLabel = new QLabel(this);
    mMessageLabel->setAlignment(Qt::AlignLeft);
    mMessageLabel->setIndent(4);
    mMessageLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    
    mClockLabel = new QLabel("00:00", this);
    mClockLabel->setAlignment(Qt::AlignHCenter);
    mClockLabel->setMinimumSize(mClockLabel->sizeHint());
    mClockLabel->setMinimumWidth(statusBar()->fontMetrics().horizontalAdvance(" 000:000 "));
    mClockLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    
    mCheatLabel = new QLabel("0(0)", this);
    mCheatLabel->setAlignment(Qt::AlignHCenter);
    mCheatLabel->setMinimumSize(mCheatLabel->sizeHint());
    mCheatLabel->setMinimumWidth(statusBar()->fontMetrics().horizontalAdvance(" 0(0) "));
    mCheatLabel->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    
    statusBar()->addWidget(mFlagCountLabel, 0);
    statusBar()->addWidget(mMessageLabel, 1);
    statusBar()->addWidget(mClockLabel, 0);
    statusBar()->addWidget(mCheatLabel, 0);
    if (!msSettings::mNumberOfAllowedOopsies) mCheatLabel->setVisible(false);
    statusBar()->setSizeGripEnabled(false);
}

void msWindow::customSizeSelected(int result)
{
    if (result == 1) {
        msSettings::mNumYCells = mCustomSizeDialog->rows();
        msSettings::mNumXCells = mCustomSizeDialog->columns();
        msSettings::mNumberOfMines = mCustomSizeDialog->mines();
        
        //
        //  In case the user selects the same size as one
        //  of the predefines sizes ...
        //
        if (msSettings::mNumXCells == msSettings::mLargeSizeParameters[0] &&
            msSettings::mNumYCells == msSettings::mLargeSizeParameters[1] &&
            msSettings::mNumberOfMines == msSettings::mLargeSizeParameters[2])
            msSettings::mCurrentGameSize = msSettings::msGameSize::large;
        else if (msSettings::mNumXCells == msSettings::mMediumSizeParameters[0] &&
                 msSettings::mNumYCells == msSettings::mMediumSizeParameters[1] &&
                 msSettings::mNumberOfMines == msSettings::mMediumSizeParameters[2])
            msSettings::mCurrentGameSize = msSettings::msGameSize::medium;
        else if (msSettings::mNumXCells == msSettings::mSmallSizeParameters[0] &&
                 msSettings::mNumYCells == msSettings::mSmallSizeParameters[1] &&
                 msSettings::mNumberOfMines == msSettings::mSmallSizeParameters[2])
            msSettings::mCurrentGameSize = msSettings::msGameSize::small;
        else
            msSettings::mCurrentGameSize = msSettings::msGameSize::custom;
        
        startGame();
    }
}

bool msWindow::eventFilter(QObject *obj, QEvent *event)
{
    //
    // If method retuns true, the event stops here otherwise its
    // passed down. Here we catch all dispatched events from the
    // image scene (canvas).
    //
    QGraphicsSceneMouseEvent *mevt;

    //
    //  X11/Linux hack: X11 doesn't know the size of the window
    //  frame until it is actually visible. resizeMainWindow()
    //  has no way of setting the sizes properly until late
    //  when things are already up and running. This is an
    //  issue only at the very start. This is the fix:
    //
#if defined (Q_OS_LINUX) || defined (Q_OS_UNIX)
    static bool x11FixOnce = true;
    if (x11FixOnce && event->type() == QEvent::WindowActivate) {
        resizeMainWindow();
        x11FixOnce = false;
    }
#endif
    
    if (obj == mGraphicsScene && mGameIsOn) {
        
        //
        //  Give animation time to finish if runnig
        //
        if (mAnimation->state() == QAbstractAnimation::Running) {
            return QObject::eventFilter(obj, event);
        }

        if (event->type() == QEvent::GraphicsSceneMousePress || event->type() == QEvent::GraphicsSceneMouseDoubleClick) {

            //
            //  Start clock at first mouse click
            //
            if (mStartTime == 0) {
                mStartTime = time(0);
                startClock();
            }

            //
            //   Get position and thus the referring cell
            //
            mevt = dynamic_cast<QGraphicsSceneMouseEvent*>(event);
            QPointF mpos = mevt->scenePos();
            unsigned int ix = static_cast<int>(mpos.x())/msSettings::mCellSize;
            unsigned int iy = static_cast<int>(mpos.y())/msSettings::mCellSize;

            //
            //   Check if ix and iy are valid. If so get the
            //   pointer to the referring cell.
            //
            msCell *cell = 0;
            if (ix < msSettings::mNumXCells && iy < msSettings::mNumYCells )
                cell = mCellArray[ix][iy];
            else
                return QObject::eventFilter(obj, event);
            
            //
            //   Right button - used to flag/unflag mines
            //
            if (mevt->button() == Qt::RightButton) {
                if (cell->isRevealed()) return false;  // nothing to do
                
                // switch from flag to try-out (or unflag)
                if (cell->isFlaggedAsMine()) {
                    if (msSettings::mAllowQuestionableFlag) mCellArray[ix][iy]->tryOutText()->show();
                    cell->setIsFlaggedAsMine(false);
                    mNumberOfFlaggedMines--;
                    updateFlaggedMineCounter();
                }
                // remove try-out
                else if (cell->tryOutText()->isVisible()) {
                    cell->tryOutText()->hide();
                }
                // set mine flag
                else {
                    cell->setIsFlaggedAsMine(true);
                    mNumberOfFlaggedMines++;
                    updateFlaggedMineCounter();
                }
            }
            //
            //   Left button - used to reveal a cell
            //
            else if (mevt->button() == Qt::LeftButton && !cell->isFlaggedAsMine()) {
                
                if (mCellArray[ix][iy]->isMine() && mOopsiesLeft && msSettings::mNumberOfAllowedOopsies > 0) {
                    startOopsAnimation(mpos);
                    mOopsiesLeft--;
                    return false;
                }
                
                if (cell->flagSvg()->isVisible()) {
                    mNumberOfFlaggedMines--;
                    updateFlaggedMineCounter();
                }
                cell->setIsRevealed(true);
                cell->waterRectangle()->hide();
                cell->flagSvg()->hide();
                cell->tryOutText()->hide();
                if (mCellArray[ix][iy]->isMine()) {
                    QBrush brush(Qt::red);
                    cell->frameRectangle()->setBrush(brush);
                    cell->frameRectangle()->setZValue(-10);
                    update();
                    endGame(false);
                    return false;
                }
                else {
                    if (cell->numberOfNeighbors() == 0) clearObvious(ix, iy);
                }
            }
            if (isGameWon()) endGame(true);
            return true;
        }
        else
            return false;
    }
    return QObject::eventFilter(obj, event);  // standard event processing
}

void msWindow::endGame(bool won)
{
    mGameIsOn = false;
    mEndTime = time(0);
    time_t duration = mEndTime-mStartTime;
    stopClock();

    if (won) {
        mMessageLabel->setText("You won the game");
        
        bool tooManyCheats = msSettings::mNumberOfAllowedOopsies > 0 &&
                             mOopsiesLeft < (msSettings::mNumberOfAllowedOopsies-1);

        if (!tooManyCheats && mHighscoreList->isHighscore(duration, msSettings::mCurrentGameSize)) {
            bool Ok;
            QString name = QInputDialog::getText(this, "MineSweeper - Enter High Score List",
                                                 "Congratulations!\nYou entered the high score list.\n\nEnter your name:",
                                                 QLineEdit::Normal, msSettings::mLastPlayer, &Ok, Qt::Sheet);
            if (Ok) {
                msSettings::mLastPlayer = name;
                msHighscoreEntry entry(name, mEndTime, duration, msSettings::mCurrentGameSize, true);
                mHighscoreList->insertEntry(entry);
                mHighscoreList->syncAndShow();
            }
        }
        else {
            QMessageBox msgBox(this);
            msgBox.setText("Congratulations, you won!");
            QString str = QString("You completed the game in %1 min and %2 sec.").arg(duration/60).arg(duration%60);
            msgBox.setInformativeText(str);
            msgBox.setStandardButtons(QMessageBox::Ok);
            msgBox.setDefaultButton(QMessageBox::Ok);
            msgBox.setWindowFlags(Qt::Sheet);
            (void) msgBox.exec();
        }
        return;
    }

    mMessageLabel->setText("Game over");
    
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            
            //
            //   Not discovered mines, just show the mine by removing the water item
            //   and the try out flag
            //
            if (mCellArray[ix][iy]->isMine() && ! mCellArray[ix][iy]->flagSvg()->isVisible()) {
                mCellArray[ix][iy]->waterRectangle()->hide();
                mCellArray[ix][iy]->tryOutText()->hide();
            }
            //
            //   Incorrectly flagged mines
            //
            else if (!mCellArray[ix][iy]->isMine() && mCellArray[ix][iy]->flagSvg()->isVisible()) {
                mCellArray[ix][iy]->waterRectangle()->hide();
                if (mCellArray[ix][iy]->numberText()) mCellArray[ix][iy]->numberText()->hide();
                mCellArray[ix][iy]->cross()->show();
            }
        }
    }
}

void msWindow::ensureBoardFitsOnScreen()
{
    //
    //  Recalculate the maximum number of rows and columns that fit
    //  on the screen. Leave a tic of space in vertical and horizontal
    //  direction.
    //  Make sure the custom size dialog does know about these limits.
    //
    QRect maxRect = QGuiApplication::primaryScreen()->availableGeometry();
    msSettings::mMaxNumXCells = maxRect.width()/msSettings::mCellSize-1;
    msSettings::mMaxNumYCells = (maxRect.height()-statusBar()->height())/msSettings::mCellSize-1;
    mCustomSizeDialog->setMaximumRows(msSettings::mMaxNumYCells);
    mCustomSizeDialog->setMaximumColumns(msSettings::mMaxNumXCells);
    
    //
    //  Calculate the maximum cell size for the current layout
    //  that fit on the screen and let the preference dialog
    //  know about it so no one can set a cell size that would
    //  the game board exceed the size of teh screen.
    //
    unsigned int maxXCellSize = maxRect.width()/msSettings::mNumXCells;
    unsigned int maxYCellSize = (maxRect.height()-statusBar()->height())/msSettings::mNumYCells;
    mPreferencesDialog->setMaxCellSize(min(maxXCellSize, maxYCellSize));
}

void msWindow::finish()
{
    QApplication::exit(0);
}

void msWindow::finishOopsAnimation()
{
    mOopsPixmap->setVisible(true);
    mGraphicsScene->removeItem(mOopsPixmap);
}

void msWindow::help()
{
    if (mHelpBrowser)
        mHelpBrowser->show();
    else
        QMessageBox::warning( 0, "MineSweeper",
                             tr("Sorry no help available.\n"
                                "Documentation files are missing. "
                                "Check your installation and reinstall if necessary."));
}

bool msWindow::isGameWon() const
{
    if (mNumberOfFlaggedMines != msSettings::mNumberOfMines) return false;

    unsigned int numberOfFoundMines = 0;
    unsigned int numberOfRevealedFields = 0;
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            msCell *cell = mCellArray[ix][iy];
            if (cell->isCorrectlyFlaggedAsMine()) numberOfFoundMines++;
            if (cell->isRevealed()) numberOfRevealedFields++;
        }
    }
    return (numberOfFoundMines == msSettings::mNumberOfMines &&
            msSettings::numberOfCells() == numberOfRevealedFields+numberOfFoundMines);
}

void msWindow::newPreferences(int val)
{
    if (val != 1) return;  // Preference Dialog exited w/o pressing OK
    
    //
    //  Colors
    //
    mGraphicsScene->setBackgroundBrush(msSettings::mBackgroundColor);
    QBrush brush(msSettings::mWaterColor);
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            mCellArray[ix][iy]->waterRectangle()->setBrush(brush);
        }
    }
    
    //
    //  New Mines and Flags by defining a new renderer.
    //  Note: the QSvgRenderer::load() function does not
    //  work. The aspect ratio is altered for reasons
    //  unknown. Apparently things are not fully reset
    //  in load(). Creating a new one does the job.
    //
    delete mFlagSvgRenderer;
    delete mMineSvgRenderer;
    delete mTryOutSvgRenderer;
    for (auto entry : mNumberSvgRenderer) delete entry.second;
    mFlagSvgRenderer = new QSvgRenderer(msSettings::flagFileName());
    mMineSvgRenderer = new QSvgRenderer(msSettings::mineFileName());
    mTryOutSvgRenderer = new QSvgRenderer(msSettings::tryOutSymbolFileName());
    const int maxNeighbors = 8;
    for (int i=1; i<=maxNeighbors; i++)
        mNumberSvgRenderer[i] = new QSvgRenderer(msSettings::numberFileName(i));
    
    //
    //  Redo the scaling (cell size might have changed)
    //
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            
            mCellArray[ix][iy]->waterRectangle()->setRect(ix*msSettings::mCellSize,
                                                          iy*msSettings::mCellSize,
                                                          msSettings::mCellSize,
                                                          msSettings::mCellSize);
            
            mCellArray[ix][iy]->frameRectangle()->setRect(ix*msSettings::mCellSize,
                                                          iy*msSettings::mCellSize,
                                                          msSettings::mCellSize,
                                                          msSettings::mCellSize);
            
            QPolygonF polygon;
            double x = ix*msSettings::mCellSize;
            double y = iy*msSettings::mCellSize;
            polygon << QPointF(x, y) << QPointF(x+msSettings::mCellSize, y+msSettings::mCellSize)
            << QPointF(x+msSettings::mCellSize/2.,y+msSettings::mCellSize/2.) << QPointF(x+msSettings::mCellSize,y)
            << QPointF(x,y+msSettings::mCellSize) << QPointF(x+msSettings::mCellSize/2.,y+msSettings::mCellSize/2.)
            << QPointF(x+msSettings::mCellSize, y+msSettings::mCellSize);  // need to get closed
            mCellArray[ix][iy]->cross()->setPolygon(polygon);
            QPen crossPen(Qt::red);
            crossPen.setWidthF(2.*msSettings::mCellSize/msSettings::mMaxCellSize); // scale width with cell size
            mCellArray[ix][iy]->cross()->setPen(crossPen);
            
            if (mCellArray[ix][iy]->mineSvg()) {
                mCellArray[ix][iy]->mineSvg()->setSharedRenderer(mMineSvgRenderer);
                positionGraphicsItem(mCellArray[ix][iy]->mineSvg(), ix, iy);
            }
            
            mCellArray[ix][iy]->flagSvg()->setSharedRenderer(mFlagSvgRenderer);
            positionGraphicsItem(mCellArray[ix][iy]->flagSvg(), ix, iy);
            
            mCellArray[ix][iy]->tryOutText()->setSharedRenderer(mTryOutSvgRenderer);
            positionGraphicsItem(mCellArray[ix][iy]->tryOutText(), ix, iy, 0.8);
            
            int n = mCellArray[ix][iy]->numberOfNeighbors();
            if (n) {
                mCellArray[ix][iy]->numberText()->setSharedRenderer(mNumberSvgRenderer[n]);
                positionGraphicsItem(mCellArray[ix][iy]->numberText(), ix, iy, 0.8);
            }
        }
    }
    
    //
    //  Resize main window for a snug fit
    //
    resizeMainWindow();
    
    //
    //  Game options
    //
    if (mOopsiesLeft == 0) mOopsiesLeft = msSettings::mNumberOfAllowedOopsies;
    mCheatLabel->setVisible(msSettings::mNumberOfAllowedOopsies);
    mCheatLabel->setText(QString(" %1(%2) ").arg(msSettings::mNumberOfAllowedOopsies-mOopsiesLeft).arg(msSettings::mNumberOfAllowedOopsies));
}

void msWindow::positionGraphicsItem(QGraphicsSvgItem* item, int ix, int iy, double magnification)
{
    if (!item) return;
    item->setScale(1);
    QRectF rect = item->boundingRect();
    double scale = 1;
    if (msSettings::mCellSize/rect.height()<scale) scale = msSettings::mCellSize/rect.height();
    if (msSettings::mCellSize/rect.width()<scale) scale = msSettings::mCellSize/rect.width();
    scale *= 0.8*magnification;
    item->setScale(scale);
    double dx = (msSettings::mCellSize-scale*rect.width())/2.;
    double dy = (msSettings::mCellSize-scale*rect.height())/2.;
    item->setPos(QPointF((ix*msSettings::mCellSize)+dx, (iy*msSettings::mCellSize)+dy));
}

unsigned int msWindow::randomNumber(unsigned int b)
{
    uniform_int_distribution<unsigned int> distribution(0,b-1);
    return distribution(mRandomEngine);
}

void msWindow::resetClock()
{
    mClockLabel->setText("00:00");
}

void msWindow::resizeMainWindow()
{
    QSize s(msSettings::mNumXCells*msSettings::mCellSize,  msSettings::mNumYCells*msSettings::mCellSize);
    //PR(msSettings::mNumXCells*msSettings::mCellSize);
    //PR(msSettings::mNumYCells*msSettings::mCellSize);
    mGraphicsView->scene()->setSceneRect(0, 0, s.width(), s.height());
    int borderWidth  = frameGeometry().width() - width();
    int borderHeight = frameGeometry().height() - height();
    QSize fs(s.width()+borderWidth+1, s.height()+borderHeight+1);
#if defined (Q_OS_LINUX) 
    //
    //  X11 fix for incorrect frameGeometry (see also eventFilter())
    //
    int x11TitleBarHeight = QApplication::style()->pixelMetric(QStyle::PM_TitleBarHeight);
    fs.setHeight(fs.height()+x11TitleBarHeight);
    // if (!isVisible()) fs.setHeight(fs.height()+ mMenuBar->height());
#elif defined(Q_OS_WIN)
    fs.setHeight(fs.height() + 2);  // avoid slight wiggle
    fs.setWidth(fs.width() - borderWidth);  // avoid slight wiggle
#endif
    //PR(fs.width());
    //PR(fs.height());
    resize(fs);
    setFixedSize(fs);
    update();
}

void msWindow::setSmallSizeAndStartNew()
{
    msSettings::mNumXCells=msSettings::mSmallSizeParameters[0];
    msSettings::mNumYCells=msSettings::mSmallSizeParameters[1];
    msSettings::mNumberOfMines=msSettings::mSmallSizeParameters[2];
    msSettings::mCurrentGameSize = msSettings::msGameSize::small;
    startGame();
}

void msWindow::setMediumSizeAndStartNew()
{
    msSettings::mNumXCells=msSettings::mMediumSizeParameters[0];
    msSettings::mNumYCells=msSettings::mMediumSizeParameters[1];
    msSettings::mNumberOfMines=msSettings::mMediumSizeParameters[2];
    msSettings::mCurrentGameSize = msSettings::msGameSize::medium;
    startGame();
}

void msWindow::setLargeSizeAndStartNew()
{
    msSettings::mNumXCells=msSettings::mLargeSizeParameters[0];
    msSettings::mNumYCells=msSettings::mLargeSizeParameters[1];
    msSettings::mNumberOfMines=msSettings::mLargeSizeParameters[2];
    msSettings::mCurrentGameSize = msSettings::msGameSize::large;
    startGame();
}

void msWindow::setCustomSize()
{
    mCustomSizeDialog->setRows(msSettings::mNumYCells);
    mCustomSizeDialog->setColumns(msSettings::mNumXCells);
    mCustomSizeDialog->setMines(msSettings::mNumberOfMines);
    mCustomSizeDialog->show();
}

void msWindow::showPreferences()
{
    mPreferencesDialog->show();
}

void msWindow::showHighscores()
{
    mHighscoreList->syncAndShow();
}

void msWindow::startClock()
{
    resetClock();
    mClockUpdateTimer->start(1000);
}

void msWindow::startGame()
{
    //
    //  Clear old game
    //  1. Remove all items from graphics scene
    //  2. Delete all cells
    //  3. Clear array holding cells
    //
    QList<QGraphicsItem*> listOfOld = mGraphicsScene->items();
    for (auto p: listOfOld) mGraphicsScene->removeItem(p);
    
    for (unsigned int k=0; k < mCellArray.size(); k++) {
        for (unsigned int j=0; j < mCellArray[k].size(); j++) {
            delete mCellArray[k][j];
        }
        mCellArray[k].clear();
    }
    mCellArray.clear();
    mNoMinesNoNeighborsCellArray.clear();
    mGraphicsScene->update();
    
    //
    //  Resize cell array to correct size
    //
    mCellArray.resize(msSettings::mNumXCells);
    for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
        mCellArray[ix].resize(msSettings::mNumYCells);
    }
    
    //
    //  Define and set limits for mCustomSizeDialog
    //  and mPreferenceDialog.
    //
    ensureBoardFitsOnScreen();
    
    //
    //  Create cells and assign xy position
    //
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            mCellArray[ix][iy] = new msCell;
            mCellArray[ix][iy]->setXY(ix, iy);
        }
    }
    
    //
    //  Generate mines, scale and position the SVG graphics properly
    //
    for (unsigned int i=0; i < msSettings::mNumberOfMines; i++) {
        unsigned int ix, iy;
        do {
            ix = randomNumber(msSettings::mNumXCells);
            iy = randomNumber(msSettings::mNumYCells);
        } while(mCellArray[ix][iy]->isMine());
        mCellArray[ix][iy]->setIsMine(true);
        QGraphicsSvgItem *item = new QGraphicsSvgItem();
        item->setSharedRenderer(mMineSvgRenderer);
        item->setFlags(QGraphicsItem::ItemClipsToShape);
        item->setCacheMode(QGraphicsItem::NoCache);
        mCellArray[ix][iy]->setMineSvg(item);
        mGraphicsScene->addItem(item);
        positionGraphicsItem(item, ix, iy);
    }
    
    //
    //  Generate "frames" rectangles
    //
    QPen pen(Qt::gray);
    
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            QGraphicsRectItem *item = new QGraphicsRectItem(QRect(ix*msSettings::mCellSize,
                                                                  iy*msSettings::mCellSize,
                                                                  msSettings::mCellSize,
                                                                  msSettings::mCellSize));
            item->setPen(pen);
            mCellArray[ix][iy]->setFrameRect(item);
            mGraphicsScene->addItem(item);
        }
    }
    
    //
    //  Assign each cell in the array the number of neighboring mines
    //  and create the text item for later use.
    //  Since the metric of a character is complex and it is hard to
    //  properly center a character in a box we use QGraphicsSvgItem
    //  instead of QGraphicsTextItem.
    //
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            
            if (!mCellArray[ix][iy]->isMine()) {
                int nMinesAround = 0;
                for (int dy=-1; dy<=1; dy++) {
                    int y = iy+dy;
                    if (y < 0 || y >= static_cast<int>(msSettings::mNumYCells)) continue;
                    for (int dx=-1; dx<=1; dx++) {
                        int x = ix+dx;
                        if (x < 0 || x >= static_cast<int>(msSettings::mNumXCells)) continue;
                        if (mCellArray[x][y]->isMine()) nMinesAround++;
                    }
                }
                mCellArray[ix][iy]->setNumberOfNeighbors(nMinesAround);
                if (nMinesAround) {
                    QGraphicsSvgItem *item = new QGraphicsSvgItem;
                    item->setSharedRenderer(mNumberSvgRenderer[nMinesAround]);
                    item->setFlags(QGraphicsItem::ItemClipsToShape);
                    item->setCacheMode(QGraphicsItem::NoCache);
                    mCellArray[ix][iy]->setNumberText(item);
                    mGraphicsScene->addItem(item);
                    positionGraphicsItem(item, ix, iy, 0.8);
                    //item->setVisible(false);
                }
                else
                    mNoMinesNoNeighborsCellArray.push_back(mCellArray[ix][iy]);
            }
        }
    }
    
    //
    //  Generate "water" rectangles
    //
    QBrush brush(msSettings::mWaterColor);
    
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            QGraphicsRectItem *item = new QGraphicsRectItem(QRect(ix*msSettings::mCellSize,
                                                                  iy*msSettings::mCellSize,
                                                                  msSettings::mCellSize,
                                                                  msSettings::mCellSize));
            item->setBrush(brush);
            item->setPen(pen);  // same as for frame
            mCellArray[ix][iy]->setWaterRect(item);
            mGraphicsScene->addItem(item);
        }
    }
    
    //
    //  Create signs to mark the mines, and crosses
    //
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            
            //
            //  Mine flags
            //
            QGraphicsSvgItem *item = new QGraphicsSvgItem();
            item->setSharedRenderer(mFlagSvgRenderer);
            item->setFlags(QGraphicsItem::ItemClipsToShape);
            item->setCacheMode(QGraphicsItem::NoCache);
            item->setVisible(false);
            mCellArray[ix][iy]->setFlagSvg(item);
            mGraphicsScene->addItem(item);
            positionGraphicsItem(item, ix, iy);
            
            //
            //  Crosses (used to mark incorrectly marked flags)
            //
            QPolygonF polygon;
            double x = ix*msSettings::mCellSize;
            double y = iy*msSettings::mCellSize;
            polygon << QPointF(x, y) << QPointF(x+msSettings::mCellSize, y+msSettings::mCellSize)
            << QPointF(x+msSettings::mCellSize/2.,y+msSettings::mCellSize/2.) << QPointF(x+msSettings::mCellSize,y)
            << QPointF(x,y+msSettings::mCellSize) << QPointF(x+msSettings::mCellSize/2.,y+msSettings::mCellSize/2.)
            << QPointF(x+msSettings::mCellSize, y+msSettings::mCellSize);  // need to get closed
            QGraphicsPolygonItem *poly = new QGraphicsPolygonItem(polygon);
            QPen crossPen(Qt::red);
            crossPen.setWidthF(2.*msSettings::mCellSize/msSettings::mMaxCellSize);
            poly->setPen(crossPen);
            poly->setVisible(false);
            mCellArray[ix][iy]->setCross(poly);
            mGraphicsScene->addItem(poly);
        }
    }
    
    //
    //  Create try out flags
    //
    for (unsigned int iy=0; iy < msSettings::mNumYCells; iy++) {
        for (unsigned int ix=0; ix < msSettings::mNumXCells; ix++) {
            QGraphicsSvgItem *item = new QGraphicsSvgItem();
            item->setSharedRenderer(mTryOutSvgRenderer);
            item->setFlags(QGraphicsItem::ItemClipsToShape);
            item->setCacheMode(QGraphicsItem::NoCache);
            mCellArray[ix][iy]->setTryOutText(item);
            mGraphicsScene->addItem(item);
            positionGraphicsItem(item, ix, iy, 0.8);
            item->setVisible(false);
        }
    }
    
    //
    //  Resize main window to a snug fit
    //
    resizeMainWindow();
    
    //
    //  Easy start option
    //
    if (msSettings::mEasyStart) {
        int i = randomNumber(static_cast<unsigned int>(mNoMinesNoNeighborsCellArray.size()));
        clearObvious(mNoMinesNoNeighborsCellArray[i]->x(), mNoMinesNoNeighborsCellArray[i]->y());
    }
    
    //
    //  Reset graphics scene background color (might have been newly
    //  defined via the preferences dialog)
    //
    mGraphicsScene->setBackgroundBrush(msSettings::mBackgroundColor);
    
    //
    //   Set counter, flags, and indicate cheats
    //
    mOopsiesLeft = msSettings::mNumberOfAllowedOopsies;
    mCheatLabel->setText(QString(" %1(%2) ").arg(msSettings::mNumberOfAllowedOopsies-mOopsiesLeft).arg(msSettings::mNumberOfAllowedOopsies));
    mNumberOfFlaggedMines = 0;
    mGameIsOn = true;
    mStartTime = 0; // clock starts at first click (see eventFilter))
    mInactiveTime = 0;
    resetClock();
    updateFlaggedMineCounter();
    mMessageLabel->setText("New game is on");
    update();
}

void msWindow::startOopsAnimation(const QPointF& mpos)
{
    QRectF rect = mOopsPixmap->boundingRect();
    mOopsPixmap->setPos(mpos.x()-rect.width()/2., mpos.y()-rect.height()/2);
    mOopsPixmap->setVisible(true);
    mGraphicsScene->addItem(mOopsPixmap);
    mAnimation->start();
    mCheatLabel->setText(QString(" %1(%2) ").arg(msSettings::mNumberOfAllowedOopsies-mOopsiesLeft+1).arg(msSettings::mNumberOfAllowedOopsies));
}

void msWindow::stopClock()
{
    mClockUpdateTimer->stop();
}

void msWindow::updateClock()
{
    if (!mGameIsOn) return;
    if (mGameIsOn && mStartTime == 0) return;
    mEndTime = time(0);
    long delta = mEndTime-mStartTime;
    long min = delta/60;
    long sec = delta-min*60;
    if (min < 0 && sec < 0) return;
    if (min > 60) {
        mClockLabel->setText("--:--");
        stopClock();
     }
    else {
        QTime time(0, static_cast<int>(min), static_cast<int>(sec), 0);
        if (!time.isValid()) {
            mClockLabel->setText("--:--");
            return;
        }
        mClockLabel->setText(time.toString("mm:ss"));
    }
    update();
 }

void msWindow::updateFlaggedMineCounter()
{
    char text[16];
    snprintf(text, 16, " %d/%d ", mNumberOfFlaggedMines, msSettings::mNumberOfMines);
    mFlagCountLabel->setText(text);
    mFlagCountLabel->repaint();
}

